/* file:  ponte_pericolante_semplice.c */

#ifndef _THREAD_SAFE
	#define _THREAD_SAFE
#endif

#include "printerror.h"

#include <unistd.h> 
#include <stdlib.h> 
#include <stdio.h> 
#include <stdint.h>
#include <inttypes.h>
#include <pthread.h> 

#include "DBGpthread.h"

#define NUMAUTOORARIO 4
#define NUMAUTOANTIORARIO 4
#define INDICESENSOORARIO 0
#define INDICESENSOANTIORARIO 1

/* dati da proteggere */
int bigliettoDistributore[2];
int numInAttesa[2];
int biglietto[2];
int PonteOccupato=0;

/* variabili per la sincronizzazione */
pthread_mutex_t  mutex;
pthread_mutex_t  mutexDistributori;
pthread_cond_t   condPonteLibero;

void *Auto (void *arg) 
{ 
	char Plabel[128];
	intptr_t indice;
	int myBiglietto, indiceSenso;
	char senso; /* O orario  A antiorario */

	indice=(intptr_t)arg;
	if( indice<NUMAUTOORARIO ) 
	{	senso='O'; indiceSenso=INDICESENSOORARIO; }
	else
	{	senso='A'; indiceSenso=INDICESENSOANTIORARIO; }

	sprintf(Plabel,"%cA%" PRIiPTR "", senso, indice);

	while(1) {
		DBGpthread_mutex_lock(&mutexDistributori,Plabel); 
		myBiglietto=bigliettoDistributore[indiceSenso];
		bigliettoDistributore[indiceSenso]++;
		DBGpthread_mutex_unlock(&mutexDistributori,Plabel); 

		printf("auto %s prende biglietto %i \n", Plabel, myBiglietto);
		fflush(stdout);

		DBGpthread_mutex_lock(&mutex,Plabel); 
		numInAttesa[indiceSenso]++;
		while(	( PonteOccupato ) ||
			( myBiglietto>biglietto[indiceSenso] ) ||
			( numInAttesa[indiceSenso]<numInAttesa[!indiceSenso] )
			  ||
			(
			  numInAttesa[indiceSenso]==numInAttesa[!indiceSenso]
			  &&
			  indiceSenso==INDICESENSOANTIORARIO
			)
		     ) {
			DBGpthread_cond_wait(&condPonteLibero,&mutex,Plabel);
		}
		numInAttesa[indiceSenso]--;
		PonteOccupato=1;
		biglietto[indiceSenso]++;
		printf("auto %s attraversa ponte biglietto %i \n", Plabel,
			myBiglietto );
		fflush(stdout);

		DBGpthread_mutex_unlock(&mutex,Plabel); 

		sleep(1); /* attraverso il ponte */

		DBGpthread_mutex_lock(&mutex,Plabel); 
		PonteOccupato=0;
		DBGpthread_cond_broadcast(&condPonteLibero,Plabel);
		DBGpthread_mutex_unlock(&mutex,Plabel); 

		printf("auto %s gira intorno \n", Plabel );
		fflush(stdout);

		sleep(5); /* faccio il giro attorno a Villa Inferno */
	}	

	pthread_exit(NULL); 
}


int main ( int argc, char* argv[] ) 
{ 
	pthread_t    th; 
	int  rc;
	intptr_t i;

	rc = pthread_cond_init(&condPonteLibero, NULL);
	if( rc ) PrintERROR_andExit(rc,"pthread_cond_init failed");

	rc = pthread_mutex_init(&mutex, NULL); 
	if( rc ) PrintERROR_andExit(rc,"pthread_mutex_init failed");

	rc = pthread_mutex_init(&mutexDistributori, NULL); 
	if( rc ) PrintERROR_andExit(rc,"pthread_mutex_init failed");

	for( i=0; i<2; i++ ) {
		numInAttesa[i]=0;
		bigliettoDistributore[i]=0;
		biglietto[i]=0;
	}

	for(i=0;i<NUMAUTOORARIO+NUMAUTOANTIORARIO;i++) {
		rc=pthread_create(&th,NULL,Auto,(void*)i); 
		if(rc) PrintERROR_andExit(rc,"pthread_create failed");
	}
	
	pthread_exit(NULL);
	return(0); 
} 
  
  
  
